/***************************************************************************
                          highscores.cpp  -  description
                             -------------------
    begin                : Tue Apr 5 2005
    copyright            : (C) 2005 by Thomas Friedrichsmeier
    email                : tfry@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "highscores.h"

#include <kconfig.h>
#include <kapplication.h>
#include <klocale.h>
#include <kinputdialog.h>
#include <kdialogbase.h>
#include <kmessagebox.h>

#include <qlistview.h>

#include <stdlib.h>

#include "cdp.h"

#define MAX_SCORES 5

Highscores::Highscores () {
	readHighscores ();
}

Highscores::~Highscores () {
	clearHighscores ();
}

void Highscores::showHighscores (const QString &mission) {
	if (!scores_map.count ()) {
		KMessageBox::sorry (0, i18n ("There are no highscores so far"), i18n ("No highscores"));
		return;
	}

	HighscoresDialog *d = new HighscoresDialog (this, mission);
	d->exec ();
	delete d;
}

void Highscores::addScore (const QString &mission_file, int start_level, int end_level, float score) {
	int num_levels = end_level - start_level;
	ScoreList *score_list;

// find or generate ScoreList for this mission
	QMap<QString, ScoreList *>::const_iterator it;
	it = scores_map.find (mission_file);
	if (it == scores_map.end ()) {
		score_list = new ScoreList;
		score_list->setAutoDelete (true);
		scores_map.insert (mission_file, score_list);

		KConfig *config = kapp->config ();
		config->setGroup ("Highscores");
		QStringList missions_list;
		for (it = scores_map.begin (); it != scores_map.end (); ++it) {
			missions_list.append (it.key ());
		}
		config->writeEntry ("missions", missions_list);
		config->sync ();
	} else {
		score_list = *it;
	}

// check whether score should be added to list. Note: list is always sorted
	int counter = 0;
	Highscore *hs = score_list->first ();
	while (hs && (counter < MAX_SCORES)) {
		if (num_levels > (hs->end_level - hs->start_level)) {
			hs = 0;			// abort loop
			--counter;
		} else if (num_levels == (hs->end_level - hs->start_level)) {
			if (end_level > hs->end_level) {
				hs = 0;
				--counter;
			} else if (end_level == hs->end_level) {
				if (score > hs->score) {
					hs = 0;
					--counter;
				}
			}
		}
		if (hs) {
			hs = score_list->next ();
		}
		++counter;
	}

// went out of loop without reaching max number of highscores - we need to add the current score
	if (counter < MAX_SCORES) {
		Highscore *hscore = new Highscore;
		hscore->start_level = start_level;
		hscore->end_level = end_level;
		hscore->score = score;
		hscore->name = KInputDialog::getText (i18n ("Honor to whom honor is due"), i18n ("You achieved a new highscore! Enter your name"), getenv ("USER"));
		hscore->date = QDate::currentDate ();
		score_list->insert (counter, hscore);

		saveHighscores (mission_file);

		showHighscores (mission_file);
	}
}

void Highscores::saveHighscores (const QString &mission) {
	// For debug purposes only, this should never happen
	if (!scores_map.contains (mission)) {
		cdp->debug_msg (i18n ("Error saving highscores. This should not happen. Please report as a bug"), Serious, 0);
		return;
	}

	QValueList<int> start_levels, end_levels, scores_times_100;
	QStringList names, dates;
	int counter = 0;
	ScoreList *list = scores_map[mission];
	for (Highscore *it = list->first (); it && (counter < MAX_SCORES); it = list->next ()) {
		start_levels.append (it->start_level);
		end_levels.append (it->end_level);
		scores_times_100.append (static_cast<int> (it->score * 100));
		names.append (it->name);
		dates.append (it->date.toString (Qt::ISODate));
		++counter;
	}

	KConfig *config = kapp->config ();
	config->setGroup ("Highscores::" + mission);

	config->writeEntry ("start_levels", start_levels);
	config->writeEntry ("end_levels", end_levels);
	config->writeEntry ("scores", scores_times_100);
	config->writeEntry ("names", names);
	config->writeEntry ("dates", dates);

	config->sync();
}

void Highscores::clearHighscores () {
	for (QMap<QString, ScoreList*>::iterator it = scores_map.begin (); it != scores_map.end (); ++it) {
		delete *it;		// ScoreList is set to AutoDelete, hence no need to delete individual scores
	}
	scores_map.clear ();
}

void Highscores::readHighscores () {
	clearHighscores ();
	
	KConfig *config = kapp->config ();
	config->setGroup ("Highscores");

	QStringList missions;
	missions = config->readPathListEntry ("missions");

	for (QStringList::const_iterator it = missions.begin (); it != missions.end (); ++it) {
		scores_map.insert (*it, readHighscores (*it));
	}
}

Highscores::ScoreList *Highscores::readHighscores (const QString &mission) {
	KConfig *config = kapp->config ();
	config->setGroup ("Highscores::" + mission);

	ScoreList *list = new ScoreList;
	list->setAutoDelete (true);
	QValueList<int> start_levels, end_levels, scores_times_100;
	QStringList names, dates;
	
	start_levels = config->readIntListEntry ("start_levels");
	end_levels = config->readIntListEntry ("end_levels");
	scores_times_100 = config->readIntListEntry ("scores");
	names = config->readListEntry ("names");
	dates = config->readListEntry ("dates");

	unsigned int num = start_levels.count ();
	if (!((num == end_levels.count ()) && (num == scores_times_100.count ()) && (num == names.count ()) && (num == dates.count ()))) {
		cdp->debug_msg (i18n ("Count mismatch reading highscore list"), Serious, 0);
		return list;
	}

	Highscore *score;
	for (unsigned int i=0; i < num; ++i) {
		score = new Highscore;
		score->start_level = start_levels[i];
		score->end_level = end_levels[i];
		score->score = scores_times_100[i] / 100.0;
		score->name = names[i];
		score->date = QDate::fromString (dates[i], Qt::ISODate);
		list->append (score);
	}
	return list;
}

////////////////////////////////////// BEGIN HighscoresDialog ///////////////////////////////////////////

#include <kdialogbase.h>

#include <qcombobox.h>
#include <qlistview.h>
#include <qvbox.h>

#include "taxipilot.h"

HighscoresDialog::HighscoresDialog (Highscores *parent, const QString &initial_mission) : KDialogBase (0, 0, true, i18n ("Taxipilot Hall of Fame"), KDialogBase::Ok) {
	hs = parent;

	QVBox *vbox = makeVBoxMainWidget ();

	int i = 0;
	int index = 0;
	combo = new QComboBox (vbox);
	for (QMap <QString, Highscores::ScoreList*>::const_iterator it = hs->scores_map.begin (); it != hs->scores_map.end (); ++it) {
		mission_files.append (it.key ());
		combo->insertItem (hs->tp->getMissionDescription (it.key ()).title);
		if (it.key () == initial_mission) {
			index = i;
		}
		++i;
	}

	listview = new QListView (vbox);
	listview->addColumn (i18n ("Rank"));
	listview->addColumn (i18n ("Name"));
	listview->addColumn (i18n ("Levels completed"));
	listview->addColumn (i18n ("Level reached"));
	listview->addColumn (i18n ("Score"));
	listview->addColumn (i18n ("Date"));
	listview->setSelectionMode (QListView::NoSelection);
	listview->setSorting (-1);

	combo->setCurrentItem (index);
	missionSelected (index);
	connect (combo, SIGNAL (activated (int)), this, SLOT (missionSelected (int)));
}

HighscoresDialog::~HighscoresDialog () {
}

void HighscoresDialog::missionSelected (int index) {
	QString mission_path = mission_files[index];
	listview->clear ();

	Highscores::ScoreList *list = hs->scores_map[mission_path];
	int pos = list->count () - 1;
	if (pos >= MAX_SCORES) pos = MAX_SCORES - 1;		// we may temporarily have more than MAX_SCORE scores
	for (Highscores::Highscore *score = list->at (pos); score; score = list->prev ()) {
		new QListViewItem (listview, QString::number (pos + 1) + ".", score->name, QString::number (score->end_level - score->start_level), QString::number (score->end_level + 1), QString::number (score->score, 'f', 2), score->date.toString (Qt::LocalDate));
		--pos;
	}
}

#include "highscores.moc"
